unit KnobsConversions;

interface

uses

  System.SysUtils, System.Classes, System.Math,

  KnobsUtils, Rationals;


const

  PI_2              = PI * 2;                        // Two times Pi, for angle conversions
  DELTA             = 1e-6;                          // Small value for equality comparisons

  MAX_SIGNAL        = 1e3;                           // Maximum signal level in a patch - for anti runaway
  MIN_SIGNAL        = - MAX_SIGNAL;                  // Minumum signal level in a patch - for anti runaway

  ShortestPossibleEnvelopeTime =  1;                 // In samples
  EnvelopeSmoothing            = 49;                 // envout := ( EnvelopeSmooting * OldValue + NewValue) / ( 1 + EnvelopeSmoothing);

var

  ReferenceA        : TSignal;  // The value of the reference A to be used - default 440 Hz
  NotesPerOctave    : TSignal;  // The number of notes wanted in an octave - there only is notation for 12 tho ... A .. G
  MiddleNote        : TSignal;  // The middle note to be used - default 69

  // NOTE : the following four CAN NOT BE CHANGED at wiil, some relations need to be maintained
  // for this see **1 and please use the SetSystemRate function to change the System_Rate.
  // This can not be moved to the implementation section as that will prevent inlining of
  // some functions. Also after changing these audio devices needto be re-initialized and lookuptables
  // must be regenerated.

  System_Rate       : Integer;  // 44100.0;                       // samples / s
  Slow_Decimation   : Integer;  // 4;                             // Decimation used for slow tick stuff (control rate)
  System_Rate_2     : TSignal;  // System_Rate * 2;               // samples / s
  Slow_Rate         : TSignal;  // System_Rate / SLOW_DECIMATION; // samples / s

  // END NOTE.

type

  TSample         = SmallInt;                                     // External value type for OS samples
  TSamples        = packed array[ 0 .. MaxInt div SizeOf( TSample) - 1] of TSample;
  PSamples        = ^TSamples;

  TSignalFunction = function( const aValue: TSignal): TSignal;


  TSignals = class
  public
    FSignals: TSignalArray;
  private
    function    GetCount : Integer;
    function    GetSignal( anIndex: Integer): TSignal;
    procedure   SetSignal( anIndex: Integer; const aValue: TSignal);
  public
    constructor Create( aCount: Integer);
    destructor  Destroy;                                                            override;
    procedure   Clear;
    procedure   AddSignal( const aValue: TSignal);
    procedure   SetSignals( aStartIndex: Integer; const aValue: TSignals);
    function    Interpolate( const anIndex: TSignal): TSignal;                    // index [0, Count-1]
    function    Convolve( const aValue: TSignal): TSignal;
  public
    property    Count                    : Integer read GetCount;
    property    Signal[ anIndex: Integer]: TSignal read GetSignal write SetSignal;
  end;

  TSignalPair = record
    Left  : TSignal;
    Right : TSignal;
  end;

  TSignalPairs = array of TSignalPair;

  TSignalQueue = class
  private
    FSignals    : TSignalPairs;
    FInp        : Integer;
    FOutp       : Integer;
    FFreeCount  : Integer;
    FFillCount  : Integer;
    FOverflows  : Integer;
    FUnderflows : Integer;
    FLastValue  : TSignalPair;
  private
    function    GetOverflows : Integer;
    function    GetUnderflows: Integer;
  public
    constructor Create( aSize: Integer);
    destructor  Destroy;                                 override;
    procedure   Clear;
    procedure   Append( const aLeft, aRight: TSignal);   overload;
    procedure   Append( const aPair: TSignalPair);       overload;
    function    GetData: TSignalPair;
  public
    property    Overflows : Integer read GetOverflows;
    property    Underflows: Integer read GetUnderflows;
  end;

  TKnobsNullConverter = class
  private
    FSearched : string;
  protected
    FName     : string;
  public
    constructor Create;                                                virtual;
    function    PosToHint   ( aPos, aSteps: Integer): string;          virtual;
    function    PosToDisplay( aPos, aSteps: Integer): string;          virtual;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;         virtual;
  public
    property    Name     : string read FName;
    property    Searched : string read FSearched write FSearched;
  end;

  TKnobsConverterClass = class of TKnobsNullConverter;

  TKnobsConverters = class( TStringList)
  private
    function    Converter( const anIndex: string): TKnobsNullConverter;
  public
    constructor Create;
    procedure   RegisterConverter( aClass: TKnobsConverterClass);
    procedure   RegisterConverters( const aClasses: array of TKnobsConverterClass);
    function    PosToHint   ( const aName: string; aPos, aSteps: Integer): string;
    function    PosToDisplay( const aName: string; aPos, aSteps: Integer): string;
    function    PosToValue  ( const aName: string; aPos, aSteps: Integer): TSignal;
  end;



  function  Remainder ( const X, Y: Integer): Integer;
  function  FlooredDiv( const X, Y: Integer): Integer;

  function  ExpAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LinAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LogAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  ExpDecay ( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LinDecay ( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LogDecay ( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;

  function  FModulo( const aValue, aDivisor: TSignal): TSignal; inline;
  function  RangeMap( const aValue, anInLow, anInHigh, anOutLow, anOutHigh: TSignal): TSignal; inline;
  function  AlmostEqual( const aValue1, aValue2: TSignal; const aDelta: TSignal = DELTA): Boolean; inline;
  function  Normalize( const aValue : TSignal): TSignal; inline;
  procedure PNormalize( var aValue : TSignal); inline;
  function  AlmostZero( const aValue : TSignal; const aDelta: TSignal = DELTA): Boolean; inline;

  function  LogicToSignal             ( const aValue : Boolean): TSignal; inline;
  function  SignalToLogic             ( const aValue : TSignal): Boolean; inline;
  function  SignalToMute              ( const aValue : TSignal): TSignal; inline;
  function  SignalToInt               ( const aValue : TSignal): Integer; inline;
  function  SignalTodB                ( const aValue : TSignal): TSignal; inline;

  function  NoteNumberToUnits         ( const aValue : TSignal): TSignal; inline;
  function  UnitsToNoteNumber         ( const aValue : TSignal): TSignal; inline;
  function  UnitsToFrequency          ( const aValue : TSignal): TSignal; inline;
  function  UnitsToFrequencyLin       ( const aValue : TSignal): TSignal; inline;
  function  FrequencyToUnits          ( const aValue : TSignal): TSignal; inline;
  function  FrequencyToPhaseDelta     ( const aValue : TSignal): TSignal; inline;
  function  UnitsToPhaseDelta         ( const aValue : TSignal): TSignal; inline;
  function  UnitsToPhaseDeltaLin      ( const aValue : TSignal): TSignal; inline;
  function  SampleCountToTime         ( const aValue : TSignal): TSignal; inline;
  function  TimeToSampleCount         ( const aValue : TSignal): TSignal; inline;

  function  UnitsToLfoFastFrequency   ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoMediumFrequency ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoSlowFrequency   ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoBPMFrequency    ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoLinFrequency    ( const aValue : TSignal): TSignal; inline;
  function  FrequencyToLfoPhaseDelta  ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoFastPhaseDelta  ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoMediumPhaseDelta( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoSlowPhaseDelta  ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoBPMPhaseDelta   ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoLinPhaseDelta   ( const aValue : TSignal): TSignal; inline;
  function  SlowTimeToSampleCount     ( const aValue : TSignal): TSignal; inline;

  function  UnitsToSvfFrequency       ( const aValue : TSignal): TSignal;
  function  UnitsToSine               ( const aValue : TSignal): TSignal;

  function  NoteNameToUnits( const aNoteName : string; const aDefault: TSignal): TSignal;
  function  UnitsToNoteName( const aValue: TSignal; Compact: Boolean = True): string;
  function  IsNoteName     ( const aNoteName: string): Boolean;

  function  SignalToStr ( const aValue: TSignal): string;

  function  TimeToStr   ( const aValue: TSignal): string;
  function  FreqToStr   ( const aValue: TSignal): string;
  function  FreqToBPMStr( const aValue: TSignal): string;

  procedure RegisterConverters( const aClasses: array of TKnobsConverterClass);
  function  Converters: TKnobsConverters;


  function  LookupFrequency          ( const aValue: TSignal): TSignal;
  function  LookupFrequencyLin       ( const aValue: TSignal): TSignal;
  function  LookupPhaseDelta         ( const aValue: TSignal): TSignal;
  function  LookupPhaseDeltaLin      ( const aValue: TSignal): TSignal;
  function  LookupSvfFrequency       ( const aValue: TSignal): TSignal;
  function  LookupLfoSlowFrequency   ( const aValue: TSignal): TSignal;
  function  LookupLfoSlowPhaseDelta  ( const aValue: TSignal): TSignal;
  function  LookupLfoMediumFrequency ( const aValue: TSignal): TSignal;
  function  LookupLfoMediumPhaseDelta( const aValue: TSignal): TSignal;
  function  LookupLfoFastFrequency   ( const aValue: TSignal): TSignal;
  function  LookupLfoFastPhaseDelta  ( const aValue: TSignal): TSignal;
  function  LookupLfoBPMFrequency    ( const aValue: TSignal): TSignal;
  function  LookupLfoBPMPhaseDelta   ( const aValue: TSignal): TSignal;
  function  LookupLfoLinFrequency    ( const aValue: TSignal): TSignal;
  function  LookupLfoLinPhaseDelta   ( const aValue: TSignal): TSignal;
  function  LookupSine               ( const aValue: TSignal): TSignal;


  procedure SetSystemRate( aValue: Integer);



implementation



var

  LocConverters : TKnobsConverters;



{ ========
  TSignals = class
  public
    FSignals: TSignalArray;
  public
    property    Count                    : Integer read GetCount;
    property    Signal[ anIndex: Integer]: TSignal read GetSignal write SetSignal;
  private
}

    function    TSignals.GetCount: Integer;
    begin
      Result := Length( FSignals);
    end;

    function    TSignals.GetSignal( anIndex: Integer): TSignal;
    begin
      Result := FSignals[ anIndex];
    end;

    procedure   TSignals.SetSignal( anIndex: Integer; const aValue: TSignal);
    begin
      FSignals[ anIndex] := aValue;
    end;

//  public

    constructor TSignals.Create( aCount: Integer); // overload;
    begin
      inherited Create;
      SetLength( FSignals, aCount);
    end;

    destructor  TSignals.Destroy; // override;
    begin
      Clear;
      inherited;
    end;

    procedure   TSignals.Clear;
    begin
      SetLength( FSignals, 0);
    end;

    procedure   TSignals.AddSignal( const aValue: TSignal);
    begin
      SetLength( FSignals, Count + 1);
      FSignals[ Count - 1] := aValue;
    end;

    procedure   TSignals.SetSignals( aStartIndex: Integer; const aValue: TSignals);
    var
      i : Integer;
    begin
      if Assigned( aValue)
      then begin
        SetLength( FSignals, aStartIndex + aValue.Count);
        for i := aStartIndex to Count - 1
        do Signal[ i] := aValue.Signal[ i - aStartIndex];
      end;
    end;

    function    TSignals.Interpolate( const anIndex: TSignal): TSignal;
    var
      Lower   : Integer;
      Higher  : Integer;
      DLower  : TSignal;
      DHigher : TSignal;
    begin
      // todo: this can not be right, can it ...
      Lower  := Floor( anIndex);
      DLower := anIndex - Lower;
      if Lower < 0
      then Lower := Count - 1;
      Higher := Ceil( anIndex);
      // if Higher = Lower
      // then Higher := Lower + 1;
      DHigher := Higher - anIndex;
      if Higher > Count - 1
      then Higher := 0;
      Result := DHigher * FSignals[ Lower] + DLower * FSignals[ Higher];
    end;

    function    TSignals.Convolve( const aValue: TSignal): TSignal;
    var
      i : Integer;
      v : TSignal;
    begin
      Result := 0;
      v      := Normalize( aValue);
      for i := 0 to Count - 1
      do Result := Result + v * Signal[ i];
    end;


{ ========
  TSignalPair = record
    Left  : TSignal;
    Right : TSignal;
  end;

  TSignalPairs = array of TSignalPair;

  TSignalQueue = class
  private
    FSignals    : TSignalPairs;
    FInp        : Integer;
    FOutp       : Integer;
    FFreeCount  : Integer;
    FFillCount  : Integer;
    FOverflows  : Integer;
    FUnderflows : Integer;
    FLastValue  : TSignalPair;
  private
}

    function    TSignalQueue.GetOverflows: Integer;
    begin
      Result := FOverflows;
      FOverflows := FOverflows - Result;
    end;

    function    TSignalQueue.GetUnderflows: Integer;
    begin
      Result := FUnderflows;
      FUnderflows := FUnderflows - Result;
    end;

//  public

    constructor TSignalQueue.Create( aSize: Integer);
    begin
      inherited Create;
      SetLength( FSignals, aSize);
      Clear;
    end;

    destructor  TSignalQueue.Destroy; // override;
    begin
      Clear;
      FFreeCount := 0;
      FFillCount := 0;
      inherited;
    end;

    procedure   TSignalQueue.Clear;
    begin
      FFillCount       := 0;
      FFreeCount       := 0;
      FInp             := 0;
      FOutp            := 0;
      FUnderflows      := 0;
      FOverflows       := 0;
      FLastValue.Left  := 0;
      FLastValue.Right := 0;
      FFreeCount       := Length( FSignals);
    end;

    procedure   TSignalQueue.Append( const aLeft, aRight: TSignal); // overload;
    var
      aPair: TSignalPair;
    begin
      aPair.Left  := aLeft;
      aPair.Right := aRight;
      Append( aPair);
    end;

    procedure   TSignalQueue.Append( const aPair: TSignalPair); // overload;
    begin
      if FFreeCount > 0
      then begin
        Dec( FFreeCount);
        FSignals[ FInp] := aPair;
        FInp := ( FInp + 1) mod Length( FSignals);
        Inc( FFillCount);
      end
      else {$Q-} Inc( FOverflows) {$Q+};
    end;

    function    TSignalQueue.GetData: TSignalPair;
    begin
      if FFillCount > 0
      then begin
        Dec( FFillCount);
        FLastValue := FSignals[ FOutp];
        FOutp      := ( FOutp + 1) mod Length( FSignals);
        Inc( FFreeCount);
      end
      else {$Q-} Inc( FUnderflows) {$Q+};
      Result := FLastValue;
    end;




  function  Remainder( const X, Y: Integer): Integer;
  begin
    Result := ( X mod Y + Y) mod Y;
  end;

  function  FlooredDiv( const X, Y: Integer): Integer;
  begin
    Result := X div Y;
    if Result < 0
    then Dec( Result);
  end;

  function  ExpAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 ^ x - 1  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := ( Power( 2, 10 * aTim) - 1) / 1023;
    Result   := Normalize(( 1 - Unscaled) * aStartValue + Unscaled * anEndValue);
  end;

  function  LinAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = x  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := aTim;
    Result   := Normalize(( 1 - Unscaled) * aStartValue + Unscaled * anEndValue);
  end;

  function  LogAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 - 2 ^ ( 1 - x)  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := ( 1024 - Power( 2, 10 - 10 * aTim)) / 1023;
    Result   := Normalize(( 1 - Unscaled) * aStartValue + Unscaled * anEndValue);
  end;

  function  ExpDecay( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 ^ ( 1 - x) - 1  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := ( Power( 2, 10 - 10 * aTim) - 1) / 1023;
    Result   := Normalize(( 1 - Unscaled) * anEndValue + Unscaled * aStartValue);
  end;

  function  LinDecay( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 1 - x  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := 1 - aTim;
    Result   := Normalize(( 1 - Unscaled) * anEndValue + Unscaled * aStartValue);
  end;

  function  LogDecay( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 - 2 ^ x  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := ( 1024 - Power( 2, 10 * aTim)) / 1023;
    Result   := Normalize(( 1 - Unscaled) * anEndValue + Unscaled * aStartValue);
  end;

  function FModulo( const aValue, aDivisor: TSignal): TSignal; inline;
  begin
    Result := aValue - ( Int( aValue / aDivisor) * aDivisor);
  end;

  function  Normalize( const aValue: TSignal): TSignal; inline;
  begin
    if Abs( aValue) < 1e-20
    then Result := 0
    else Result := aValue;
  end;

  procedure PNormalize( var aValue : TSignal); inline;
  begin
    if Abs( aValue) < 1e-20
    then aValue := 0;
  end;

  function  RangeMap( const aValue, anInLow, anInHigh, anOutLow, anOutHigh: TSignal): TSignal; inline;
  begin
    Result := anOutLow + (( aValue - anInLow) / ( anInHigh - anInLow)) * ( anOutHigh - anOutLow);
  end;

  function  AlmostEqual( const aValue1, aValue2: TSignal; const aDelta: TSignal = DELTA): Boolean;  inline;
  begin
    Result := Abs( aValue1 - aValue2) < aDelta;
  end;

  function  AlmostZero( const aValue: TSignal; const aDelta: TSignal = DELTA): Boolean; inline;
  begin
    Result := Abs( aValue) < aDelta;
  end;

  function  LogicToSignal( const aValue: Boolean): TSignal; inline;
  begin
    if aValue
    then Result := 1
    else Result := -1;
  end;

  function  SignalToLogic( const aValue: TSignal): Boolean; inline;
  begin
    Result := aValue > 0;
  end;

  function  SignalToMute( const aValue  : TSignal): TSignal; inline;
  begin
    Result := 1 - aValue;
  end;

  function  SignalToInt( const aValue : TSignal): Integer; inline;
  begin
    Result := Round( aValue + 0.4999);
  end;

  function  SignalTodB( const aValue : TSignal): TSignal; inline;
  begin
    if aValue <= 1e-7
    then Result := -140
    else Result := 20 * Log10( aValue);
  end;


  function  UnitsToNoteNumber( const aValue: TSignal): TSignal; inline;
  begin
    Result := 128 * aValue;
  end;

  function  NoteNumberToUnits( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue / 128;
  end;

  function  UnitsToFrequency( const aValue: TSignal): TSignal; inline;
  // such that note Nr 69 maps to A 440 Hz
  // see: http://www.phys.unsw.edu.au/jw/notes.html
  // freq = ( 2 ^ ( notenr - 69) / 12) * 440
  // Result := Power( 2, ( aValue - 69) / 12) * ReferenceA;
  begin
    Result := Power( 2, ( UnitsToNoteNumber( aValue) - MiddleNote) / NotesPerOctave) * ReferenceA;
  end;

  function  UnitsToFrequencyLin( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * 880;
  end;

  function  FrequencyToUnits( const aValue: TSignal): TSignal; inline;
  // see: http://www.phys.unsw.edu.au/jw/notes.html
  // n  = 12 * log2( f / 440) + 69
  //  Result := Round( 12 * Log2( aValue / ReferenceA)) + 69;
  begin
    Result := NoteNumberToUnits( NotesPerOctave * Log2( aValue / ReferenceA) + MiddleNote);
  end;

  function  FrequencyToPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue / System_Rate;
  end;

  function  UnitsToPhaseDelta( const aValue : TSignal): TSignal; inline;
  begin
    Result := FrequencyToPhaseDelta( UnitsToFrequency( aValue));
  end;

  function  UnitsToPhaseDeltaLin( const aValue : TSignal): TSignal; inline;
  begin
    Result := FrequencyToPhaseDelta( UnitsToFrequencyLin( aValue));
  end;

  function  SampleCountToTime( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue / System_Rate;
  end;

  function  TimeToSampleCount( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * System_Rate;
  end;

  // ------------------------------------

  function  UnitsToLfoFastFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( Power( 2, ( UnitsToNoteNumber( aValue) - 43) / 12) * 3.05);
  end;

  function  UnitsToLfoMediumFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( Power( 2, ( UnitsToNoteNumber( aValue) - 103) / 12) * 3.05);
  end;

  function  UnitsToLfoSlowFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( Power( 2, ( UnitsToNoteNumber( aValue) - 163) / 12) * 3.05);
  end;

  function  UnitsToLfoBPMFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( UnitsToNoteNumber( aValue) / 30);
  end;

  function  UnitsToLfoLinFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := aValue * 15;
  end;

  function  FrequencyToLfoPhaseDelta( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( Slow_Decimation * aValue / System_Rate);
  end;

  function  UnitsToLfoFastPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoFastFrequency( aValue));
  end;

  function  UnitsToLfoMediumPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoMediumFrequency( aValue));
  end;

  function  UnitsToLfoSlowPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoSlowFrequency( aValue));
  end;

  function  UnitsToLfoBPMPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoBPMFrequency( aValue));
  end;

  function  UnitsToLfoLinPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoLinFrequency( aValue))
  end;

  function  SlowTimeToSampleCount( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * Slow_Rate;
  end;

  function  UnitsToSvfFrequency( const aValue: TSignal): TSignal;
  begin
    Result := Normalize( 2 * Sin( PI * Min( 0.25, ( UnitsToFrequency( aValue) / System_Rate_2))));
  end;

  function  UnitsToSine( const aValue: TSignal): TSignal;
  begin
    Result := Normalize( Sin( PI * aValue));
  end;


  function  NoteNameToUnits( const aNoteName: string; const aDefault: TSignal): TSignal;
  // C[#][-]0
  var
    Base  : Integer;
    Sharp : Integer;
    Sign  : Integer;
    N     : Char;
    Shift : Integer;
    p     : Integer;
    T     : string;
    NN    : Integer;
  begin
    Shift :=  0;
    Sharp :=  0;
    Sign  :=  1;
    Base  := -1;
    if Length( aNoteName) > 0
    then begin
      N := UpCase( aNoteName[ 1]);
      case N of
        'C' : Base := 12     ;
        'D' : Base := 12 +  2;
        'E' : Base := 12 +  4;
        'F' : Base := 12 +  5;
        'G' : Base := 12 +  7;
        'A' : Base := 12 +  9;
        'B' : Base := 12 + 11;
        else  Base := -1;
      end;
      if ( Length( aNoteName) > 1) and ( Base > 0)
      then begin
        if aNoteName[ 2] = '#'
        then begin
          Sharp := 1;
          p     := 2;
        end
        else p := 1;
      end
      else p := 0;
      if ( Length( aNoteName) > p) and ( base > 0)
      then begin
        if aNoteName[ p + 1] = '-'
        then begin
          Sign := -1;
          p    := p + 1;
        end;
      end;
      T := Copy( aNoteName, p + 1, Length( aNoteName));
      if T[ Length( T)] = '#'
      then begin
        Sharp := 1;
        T     := Copy( T, 1, Length( T) - 1);
      end;
      Shift := StrToIntDef( T, 0);
    end;
    if Base > 0
    then begin
      NN     := Base + Sharp + Sign * Shift * 12;
      Result := NoteNumberToUnits( NN);
    end
    else Result := aDefault;
  end;

  function  UnitsToNoteName( const aValue: TSignal; Compact: Boolean = True): string;
  const
    Notes  : array[ 0 .. 11] Of Char = 'CCDDEFFGGAAB';
    Sharps : array[ 0 .. 11] Of Char = ' # #  # # # ';
  var
    i          : Integer;
    S          : string;
    aNoteValue : Integer;
  begin
    if IsNan( aValue)
    then aNoteValue := Round( MiddleNote)
    else aNoteValue := Round( UnitsToNoteNumber( aValue));
    i          := Remainder( aNoteValue, 12);
    S          := Sharps[ i];
    if Compact and ( S = ' ')
    then S := '';
    Result := Format( '%s%s%d', [ Notes[ i], S, ( FlooredDiv( aNoteValue, 12)) - 1], AppLocale);
  end;

  function  IsNoteName( const aNoteName: string): Boolean;
  // C[#][-]0[#]
  var
    p : Integer;
    i : Integer;
  begin
    Result := False;
    if Length( aNoteName) > 0
    then begin
      Result := CharInSet( aNoteName[ 1], ['a' .. 'g', 'A' .. 'G']);
      if Result and ( Length( aNoteName) > 1)
      then begin
        if aNoteName[ 2] = '#'
        then p := 2
        else p := 1;
      end
      else p := 0;

      if Result and ( Length( aNoteName) > p) and ( aNoteName[ p + 1] = '-')
      then p := p + 1;

      if Result
      then begin
        for i := p + 1 to Length( aNoteName)
        do begin
          if not CharInset( aNoteName[ i], [ '0' .. '9'])
          then begin
            if ( aNoteName[ i] = '#') and ( i = Length( aNoteName))
            then Break
            else begin
              Result := False;
              Break;
            end;
          end;
        end;
      end;
    end;
  end;

  function  SignalToStr( const aValue: TSignal): string;
  begin
    Result := FloatToStrF( aValue, ffGeneral, 15, 0, AppLocale);
  end;

  function  TimeToStr( const aValue: TSignal): string;
  begin
    if aValue = 0
    then Result := '0 s'
    else if aValue < 0.0001
    then Result := Format( '%.1f s', [ aValue * 1e6], AppLocale)
    else if aValue < 0.001
    then Result := Format( '%.0f s', [ aValue * 1e6], AppLocale)
    else if aValue < 0.01
    then Result := Format( '%.2f ms', [ aValue * 1e3], AppLocale)
    else if aValue < 0.1
    then Result := Format( '%.1f ms', [ aValue * 1e3], AppLocale)
    else if aValue < 1
    then Result := Format( '%.0f ms', [ aValue * 1e3], AppLocale)
    else if aValue < 10
    then Result := Format( '%.2f s' , [ aValue      ], AppLocale)
    else if aValue < 60
    then Result := Format( '%.1f s' , [ aValue      ], AppLocale)
    else if aValue < 3600
    then Result := FormatDateTime( 'nn:ss'  , aValue / ( 24 * 3600), AppLocale)
    else Result := FormatDateTime( 'h:nn:ss', aValue / ( 24 * 3600), AppLocale);
  end;

  function  FreqToStr( const aValue: TSignal): string;
  begin
    if aValue = 0
    then Result := '0 Hz'
    else if Abs( aValue) < 0.1
    then Result := TimeToStr( 1 / aValue)
    else if Abs( aValue) < 1
    then Result := Format( '%.3f Hz', [ aValue], AppLocale)
    else if Abs( aValue) < 10
    then Result := Format( '%.3f Hz', [ aValue], AppLocale)
    else if Abs( aValue) < 100
    then Result := Format( '%.2f Hz', [ aValue], AppLocale)
    else if Abs( aValue) < 1000
    then Result := Format( '%.1f Hz', [ aValue], AppLocale)
    else if Abs( aValue) < 10000
    then Result := Format( '%.2f kHz', [ aValue / 1000], AppLocale)
    else Result := Format( '%.1f kHz', [ aValue / 1000], AppLocale);
  end;

  function  FreqToBPMStr( const aValue: TSignal): string;
  begin
    Result := Format( '%.0f', [ aValue * 60]);
  end;



const

  LutRange = 8192;          // [0,1] TSignal range maps to [0,LutRange] table section
  LutSize  = 2 * LutRange;  // [0,2] TSignal range maps to [0,LutSize ] table section

type

  TLut = array[ - LutSize .. LutSize + 1] of TSignal;

var

  UnitsToFrequencyTable           : TLut;
  UnitsToFrequencyLinTable        : TLut;
  UnitsToPhaseDeltaTable          : TLut;
  UnitsToPhaseDeltaLinTable       : TLut;
  UnitsToSvfFreqTable             : TLut;
  UnitsToLfoSlowFrequencyTable    : TLut;
  UnitsToLfoSlowPhaseDeltaTable   : TLut;
  UnitsToLfoMediumFrequencyTable  : TLut;
  UnitsToLfoMediumPhaseDeltaTable : TLut;
  UnitsToLfoFastFrequencyTable    : TLut;
  UnitsToLfoFastPhaseDeltaTable   : TLut;
  UnitsToLfoBPMFrequencyTable     : TLut;
  UnitsToLfoBPMPhaseDeltaTable    : TLut;
  UnitsToLfoLinFrequencyTable     : TLut;
  UnitsToLfoLinPhaseDeltaTable    : TLut;
  UnitsToSineTable                : TLut;


  procedure ComputeTable( var aTable: TLut; const aFunc: TSignalFunction);
  var
    i : Integer;
  begin
    for i := Low( aTable) to High( aTable)
    do aTable[ i] := aFunc( i / LutRange);
  end;

  function TableLookup( const aValue: TSignal; const aTable : TLut): TSignal; inline;
  var
    aLow   : Integer;
    aVal   : TSignal;
    aDelta : TSignal;
  begin
    aVal   := Clip( aValue * LutRange, Low( aTable), High( aTable) - 1);
    aLow   := Trunc( aVal);
    aDelta := aVal - aLow;
    Result := aTable[ aLow];
    Result := Result + aDelta * ( aTable[ aLow + 1] - Result);
  end;

  procedure ComputeTables;
  begin
    ComputeTable( UnitsToFrequencyTable          , UnitsToFrequency          );
    ComputeTable( UnitsToFrequencyLinTable       , UnitsToFrequencyLin       );
    ComputeTable( UnitsToPhaseDeltaTable         , UnitsToPhaseDelta         );
    ComputeTable( UnitsToPhaseDeltaLinTable      , UnitsToPhaseDeltaLin      );
    ComputeTable( UnitsToSvfFreqTable            , UnitsToSvfFrequency       );
    ComputeTable( UnitsToLfoSlowFrequencyTable   , UnitsToLfoSlowFrequency   );
    ComputeTable( UnitsToLfoSlowPhaseDeltaTable  , UnitsToLfoSlowPhaseDelta  );
    ComputeTable( UnitsToLfoMediumFrequencyTable , UnitsToLfoMediumFrequency );
    ComputeTable( UnitsToLfoMediumPhaseDeltaTable, UnitsToLfoMediumPhaseDelta);
    ComputeTable( UnitsToLfoFastFrequencyTable   , UnitsToLfoFastFrequency   );
    ComputeTable( UnitsToLfoFastPhaseDeltaTable  , UnitsToLfoFastPhaseDelta  );
    ComputeTable( UnitsToLfoBPMFrequencyTable    , UnitsToLfoBPMFrequency    );
    ComputeTable( UnitsToLfoBPMPhaseDeltaTable   , UnitsToLfoBPMPhaseDelta   );
    ComputeTable( UnitsToLfoLinFrequencyTable    , UnitsToLfoLinFrequency    );
    ComputeTable( UnitsToLfoLinPhaseDeltaTable   , UnitsToLfoLinPhaseDelta   );
    ComputeTable( UnitsToSineTable               , UnitsToSine               );
  end;

  function  LookupFrequency          ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToFrequencyTable          ); end;
  function  LookupFrequencyLin       ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToFrequencyLinTable       ); end;
  function  LookupPhaseDelta         ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToPhaseDeltaTable         ); end;
  function  LookupPhaseDeltaLin      ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToPhaseDeltaLinTable      ); end;
  function  LookupSvfFrequency       ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToSvfFreqTable            ); end;
  function  LookupLfoSlowFrequency   ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoSlowFrequencyTable   ); end;
  function  LookupLfoSlowPhaseDelta  ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoSlowPhaseDeltaTable  ); end;
  function  LookupLfoMediumFrequency ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoMediumFrequencyTable ); end;
  function  LookupLfoMediumPhaseDelta( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoMediumPhaseDeltaTable); end;
  function  LookupLfoFastFrequency   ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoFastFrequencyTable   ); end;
  function  LookupLfoFastPhaseDelta  ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoFastPhaseDeltaTable  ); end;
  function  LookupLfoBPMFrequency    ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoBPMFrequencyTable    ); end;
  function  LookupLfoBPMPhaseDelta   ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoBPMPhaseDeltaTable   ); end;
  function  LookupLfoLinFrequency    ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoLinFrequencyTable    ); end;
  function  LookupLfoLinPhaseDelta   ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoLinPhaseDeltaTable   ); end;
  function  LookupSine               ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToSineTable               ); end;


type

  TConvUndefined = class( TKnobsNullConverter)
  public
    constructor Create;                                               override;
    function    PosToHint   ( aPos, aSteps: Integer): string;         override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;

  TKnobsBaseConverter = class( TKnobsNullConverter)
  public
    function    PosToHint   ( aPos, aSteps: Integer): string;         override;
  end;

  TConvLinear0_1 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLinear0_4 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLinearMin1_1 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLinearMin4_4 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvAdd1 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvQSteps = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvdB0_1 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvdB0_4 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvOscFreqCoarse = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvOscLinCoarse = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLfoFreqCoarseFast = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLfoFreqCoarseMedium = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLfoFreqCoarseSlow = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLfoFreqCoarseBPM = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLfoFreqCoarseLin = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvNoiseColor = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvEnvTime = class( TKnobsBaseConverter)
  public
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;

  TConvEnvTimeFast = class( TConvEnvTime)
  public
    constructor Create;                                               override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvEnvTimeMedium = class( TConvEnvTime)
  public
    constructor Create;                                               override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvEnvTimeSlow = class( TConvEnvTime)
  public
    constructor Create;                                               override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvMidiChIn = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvMidiChOut = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvNoteName = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvDelayTime = class( TKnobsBaseConverter)
  public
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;

  TConvDelayTimeShort = class( TConvDelayTime)
  public
    constructor Create;                                               override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvDelayTimeMedium = class( TConvDelayTime)
  public
    constructor Create;                                               override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvDelayTimeLong = class( TConvDelayTime)
  public
    constructor Create;                                               override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvFilterFreq = class( TConvOscFreqCoarse)
  public
    constructor Create;                                               override;
  end;

  TConvCents = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


  function  KnobRangeLin( const aPos, aStepCount: Integer; anOutLow, anOutHigh: TSignal): TSignal; inline;
  begin
    if aStepCount <= 0
    then Result := anOutLow
    else Result := anOutLow + ( aPos / aStepCount) * ( anOutHigh - anOutLow);
  end;

  function  KnobRangeLog( const aPos, aStepCount: Integer; anOutLow, anOutHigh: TSignal): TSignal; inline;
  // this is not really a log range but a mirrored exp curve rather
  var
    p : TSignal;
  begin
    if aStepCount <= 0
    then Result := anOutLow
    else begin
      p      := ( 1024 - Power( 2, 10 - 10 * ( aPos / aStepCount)) - 1) / 1023;
      Result := ( 1 - p) * anOutLow + p * anOutHigh;
    end;
  end;


  function  KnobRangeExp( const aPos, aStepCount: Integer; anOutLow, anOutHigh: TSignal): TSignal; inline;
  var
    p : TSignal;
  begin
    if aStepCount <= 0
    then Result := anOutLow
    else begin
      p      := ( Power( 2, 10 * ( aPos / aStepCount)) - 1) / 1023;
      Result := ( 1 - p) * anOutLow + p * anOutHigh;
    end;
  end;


{ ========
  TKnobsNullConverter = class
  private
    FSearched : string;
  protected
    FName     : string;
  public
    property    Name     : string read FName;
    property    Searched : string read FSearched write FSearched;
  public
}

    constructor TKnobsNullConverter.Create; // virtual;
    begin
      FName     := '';
      FSearched := '';
    end;

    function    TKnobsNullConverter.PosToHint( aPos, aSteps: Integer): string; // virtual;
    begin
      Result := IntToStr( aPos);
    end;

    function    TKnobsNullConverter.PosToDisplay( aPos, aSteps: Integer): string; // virtual;
    begin
      Result := IntToStr( aPos);
    end;

    function    TKnobsNullConverter.PosToValue( aPos, aSteps: Integer): TSignal; // virtual;
    begin
      Result := aPos;
    end;

{ ========
  TKnobsConverterUndefined = class( TKnobsNullConverter)
  public
}
    constructor TConvUndefined.Create; // override;
    begin
      FName := 'undefined converter';
    end;

    function    TConvUndefined.PosToHint( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( 'undefined %s'^M'does %s'^M'%f', [ Searched, inherited, PosToValue( aPos, aSteps)]);
    end;

    function    TConvUndefined.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( 'undefined %s does %s [%f]', [ Searched, inherited, PosToValue( aPos, aSteps)]);
    end;


{ ========
  TKnobsBaseConverter = class( TKnobsNullConverter)
  public
}

    function    TKnobsBaseConverter.PosToHint( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%s'^M''^M'%d', [ PosToDisplay( aPos, aSteps), aPos]);
    end;


{ ========
  TConvLinear0_1 = class( TKnobsBaseConverter)
  public
}

    constructor TConvLinear0_1.Create; // override;
    begin
      FName := 'Linear[0,1]';
    end;

    function    TConvLinear0_1.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.3f', [ PosToValue( aPos, aSteps)]);
    end;

    function    TConvLinear0_1.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;


{ ========
  TConvLinear0_4 = class( TKnobsBaseConverter)
  public
}

    constructor TConvLinear0_4.Create; // override;
    begin
      FName := 'Linear[0,4]';
    end;

    function    TConvLinear0_4.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.3f', [ PosToValue( aPos, aSteps)]);
    end;

    function    TConvLinear0_4.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 4);
    end;


{ ========
  TConvLineMin1_1 = class( TKnobsBaseConverter)
  public
}

    constructor TConvLinearMin1_1.Create; // override;
    begin
      FName := 'Linear[-1,1]';
    end;

    function    TConvLinearMin1_1.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.3f', [ PosToValue( aPos, aSteps)]);
    end;

    function    TConvLinearMin1_1.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, -1, 1);
    end;



{ ========
  TConvLineMin4_4 = class( TKnobsBaseConverter)
  public
}

    constructor TConvLinearMin4_4.Create; // override;
    begin
      FName := 'Linear[-4,4]';
    end;

    function    TConvLinearMin4_4.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.2f', [ PosToValue( aPos, aSteps)]);
    end;

    function    TConvLinearMin4_4.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, -4, 4);
    end;



{ ========
  TConvAdd1 = class( TKnobsBaseConverter)
  public
}

    constructor TConvAdd1.Create; // override;
    begin
      FName := 'Add1';
    end;

    function    TConvAdd1.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.0f', [ PosToValue( aPos, aSteps)]);
    end;

    function    TConvAdd1.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos + 1;
    end;


{ ========
  TConvQSteps = class( TKnobsBaseConverter)
  public
}

    constructor TConvQSteps.Create; // override;
    begin
      FName := 'QSteps';
    end;

    function    TConvQSteps.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      R : TRational;
    begin
      R := PosToValue( aPos, aSteps) / 64;
      R.Invert;
      Result := Format( '%s', [ R.AsString]);
    end;

    function    TConvQSteps.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos + 1;
    end;


{ ========
  TConvdB0_1 = class( TKnobsBaseConverter)
  public
}

    constructor TConvdB0_1.Create; // override;
    begin
      FName := 'dB[0,1]';
    end;

    function    TConvdB0_1.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := SignalTodB( aValue);
      Result := Format( '%.1f dB', [ aValue]);
    end;

    function    TConvdB0_1.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 1);
    end;



{ ========
  TConvdB0_4 = class( TKnobsBaseConverter)
  public
}

    constructor TConvdB0_4.Create; // override;
    begin
      FName := 'dB[0,4]';
    end;

    function    TConvdB0_4.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := SignalTodB( aValue);
      Result := Format( '%.1f dB', [ aValue]);
    end;

    function    TConvdB0_4.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 4);
    end;


{ ========
  TConvOscFreqCoarse = class( TKnobsBaseConverter)
  public
}

    constructor TConvOscFreqCoarse.Create; // override;
    begin
      FName := 'OscFreqCoarse';
    end;

    function    TConvOscFreqCoarse.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupFrequency( aValue);
      Result := FreqToStr( aValue);
    end;

    function    TConvOscFreqCoarse.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvOscLinCoarse = class( TKnobsBaseConverter)
  public
}

    constructor TConvOscLinCoarse.Create; // override;
    begin
      FName := 'OscLinCoarse';
    end;

    function    TConvOscLinCoarse.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupFrequencyLin( aValue);
      Result := FreqToStr( aValue);
    end;

    function    TConvOscLinCoarse.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvLfoFreqCoarse = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseFast.Create; // override;
    begin
      FName := 'LfoFreqCoarseFast';
    end;

    function    TConvLfoFreqCoarseFast.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoFastFrequency( aValue);
      Result := FreqToStr( aValue);
    end;

    function    TConvLfoFreqCoarseFast.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvLfoFreqCoarseMedium = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseMedium.Create; // override;
    begin
      FName := 'LfoFreqCoarseMedium';
    end;

    function    TConvLfoFreqCoarseMedium.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoMediumFrequency( aValue);
      Result := FreqToStr( aValue);
    end;

    function    TConvLfoFreqCoarseMedium.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvLfoFreqCoarseSlow = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseSlow.Create; // override;
    begin
      FName := 'LfoFreqCoarseSlow';
    end;

    function    TConvLfoFreqCoarseSlow.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoSlowFrequency( aValue);
      Result := FreqToStr( aValue);
    end;

    function    TConvLfoFreqCoarseSlow.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;



{ ========
  TConvLfoFreqCoarseBPM = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseBPM.Create; // override;
    begin
      FName := 'LfoFreqCoarseBPM';
    end;

    function    TConvLfoFreqCoarseBPM.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoBPMFrequency( aValue);
      Result := FreqToBPMStr( aValue);
    end;

    function    TConvLfoFreqCoarseBPM.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;



{ ========
  TConvLfoFreqCoarseLin = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseLin.Create; // override;
    begin
      FName := 'LfoFreqCoarseLin';
    end;

    function    TConvLfoFreqCoarseLin.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoLinFrequency( aValue);
      Result := FreqToStr( aValue);
    end;

    function    TConvLfoFreqCoarseLin.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;



{ ========
  TConvNoiseColor = class( TKnobsBaseConverter)
  public
}

    constructor TConvNoiseColor.Create; // override;
    begin
      FName := 'NoiseColor';
    end;

    function    TConvNoiseColor.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      if aPos = 0
      then Result := 'off'
      else if aPos = aSteps
      then Result := 'white'
      else Result := Format( '%.2f', [ PosToValue( aPos, aSteps)]);
    end;

    function    TConvNoiseColor.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;


{ ========
  TConvEnvTime = class( TKnobsBaseConverter)
  public
}

    function    TConvEnvTime.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := TimeToStr( PosToValue( aPos, aSteps));
    end;


{ ========
  TConvEnvTimeFast = class( TConvEnvTime)
  public
}

    constructor TConvEnvTimeFast.Create; // override;
    begin
      FName := 'EnvTimeFast';
    end;

    function    TConvEnvTimeFast.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0.0005, 12);
    end;



{ ========
  TConvEnvTimeMedium = class( TConvEnvTime)
  public
}

    constructor TConvEnvTimeMedium.Create; // override;
    begin
      FName := 'EnvTimeMedium';
    end;

    function    TConvEnvTimeMedium.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0.05, 1200);
    end;



{ ========
  TConvEnvTimeSlow = class( TConvEnvTime)
  public
}

    constructor TConvEnvTimeSlow.Create; // override;
    begin
      FName := 'EnvTimeSlow';
    end;

    function    TConvEnvTimeSlow.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 5, 120000);
    end;


{ ========
  TConvMidiCCIn = class( TKnobsBaseConverter)
  public
}

    constructor TConvMidiChIn.Create; // override;
    begin
      FName := 'MidiChIn';
    end;

    function    TConvMidiChIn.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      if  aPos = aSteps - 1
      then Result := 'omni'
      else Result := Format( 'ch %.0f', [ PosToValue( aPos, aSteps)]);
    end;

    function    TConvMidiChIn.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos + 1;
    end;


{ ========
  TConvMidiCCOut = class( TKnobsbaseConverter)
  public
}

    constructor TConvMidiChOut.Create; // override;
    begin
      FName := 'MidiChOut';
    end;

    function    TConvMidiChOut.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( 'ch %.0f', [ PosToValue( aPos, aSteps)]);
    end;

    function    TConvMidiChOut.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos + 1;
    end;



{ ========
  TConvNoteName = class( TKnobsBaseConverter)
  public
}

    constructor TConvNoteName.Create; // override;
    begin
      FName := 'NoteName';
    end;

    function    TConvNoteName.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := UnitsToNoteName( PosToValue( aPos, aSteps));
    end;

    function    TConvNoteName.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvDelayTime = class( TKnobsBaseConverter)
  public
}

    function    TConvDelayTime.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := TimeToStr( PosToValue( aPos, aSteps));
    end;


{ ========
  TConvDelayTimeShort = class( TKnobsBaseConverter)
  public
}

    constructor TConvDelayTimeShort.Create; // override;
    begin
      FName := 'DelayTimeShort';
    end;

    function    TConvDelayTimeShort.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 1e-4, 3e-2);
    end;



{ ========
  TConvDelayTimeMedium = class( TKnobsBaseConverter)
  public
}

    constructor TConvDelayTimeMedium.Create; // override;
    begin
      FName := 'DelayTimeMedium';
    end;

    function    TConvDelayTimeMedium.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 1e-2, 3);
    end;



{ ========
  TConvDelayTimeLong = class( TKnobsBaseConverter)
  public
}

    constructor TConvDelayTimeLong.Create; // override;
    begin
      FName := 'DelayTimeLong';
    end;

    function    TConvDelayTimeLong.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 1, 3e2);
    end;



{ ========
  TConvFilterFreq = class( TKnobsBaseConverter)
  public
}

    constructor TConvFilterFreq.Create; // override;
    begin
      FName := 'FilterFreq';
    end;


{ ========
  TConvCents = class( TKnobsBaseConverter)
  public
}

    constructor TConvCents.Create; // override;
    begin
      FName := 'Cents';
    end;

    function    TConvCents.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.0f ct', [ 50 * 64 * PosToValue( aPos, aSteps)]);
    end;

    function    TConvCents.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, -1 / 64, 1 / 64);
    end;

{ ========
  TKnobsConverters = class( TStringList)
  private
 }

    function    TKnobsConverters.Converter( const anIndex: string): TKnobsNullConverter;
    var
      p : Integer;
    begin
      p  := IndexOf( anIndex);
      if p < 0
      then begin
        p  := IndexOf( 'undefined converter');
        if p >= 0
        then (Objects[ p] as TKnobsNullConverter).Searched := anIndex;
      end;
      Result := Objects[ p] as TKnobsNullConverter;
    end;

 // public

    constructor TKnobsConverters.Create;
    begin
      inherited;
      OwnsObjects   := True;
      Duplicates    := dupError;
      Sorted        := True;
      CaseSensitive := False;
      RegisterConverters(
        [
          TKnobsNullConverter,
          TConvUndefined,
          TConvLinear0_1,
          TConvLinear0_4,
          TConvLinearMin1_1,
          TConvLinearMin4_4,
          TConvdB0_1,
          TConvdB0_4,
          TConvOscFreqCoarse,
          TConvNoiseColor,
          TConvOscLinCoarse,
          TConvLfoFreqCoarseFast,
          TConvLfoFreqCoarseMedium,
          TConvLfoFreqCoarseSlow,
          TConvLfoFreqCoarseBPM,
          TConvLfoFreqCoarseLin,
          TConvEnvTimeFast,
          TConvEnvTimeMedium,
          TConvEnvTimeSlow,
          TConvAdd1,
          TConvQSteps,
          TConvMidiChIn,
          TConvMidiChOut,
          TConvNoteName,
          TConvDelayTimeShort,
          TConvDelayTimeMedium,
          TConvDelayTimeLong,
          TConvFilterFreq,
          TConvCents
        ]
      );
    end;

    procedure   TKnobsConverters.RegisterConverter( aClass: TKnobsConverterClass);
    var
      C: TKnobsNullConverter;
    begin
      C := aClass.Create;
      AddObject( C.Name, C);
    end;

    procedure   TKnobsConverters.RegisterConverters( const aClasses: array of TKnobsConverterClass);
    var
      C: TKnobsConverterClass;
    begin
      for C in aClasses
      do RegisterConverter( C);
    end;

    function    TKnobsConverters.PosToHint( const aName: string; aPos, aSteps: Integer): string;
    begin
      Result := Converter( aName).PosToHint( aPos, aSteps);
    end;

    function    TKnobsConverters.PosToDisplay( const aName: string; aPos, aSteps: Integer): string;
    begin
      Result := Converter( aName).PosToDisplay( aPos, aSteps);
    end;

    function    TKnobsConverters.PosToValue( const aName: string; aPos, aSteps: Integer): TSignal;
    begin
      Result := Converter( aName).PosToValue( aPos, aSteps);
    end;



  procedure RegisterConverters( const aClasses: array of TKnobsConverterClass);
  begin
    LocConverters.RegisterConverters( aClasses);
  end;

  function  Converters: TKnobsConverters;
  begin
    Result := LocConverters;
  end;

  procedure CreateConverters;
  begin
    LocConverters := TKnobsConverters.Create;
  end;

  procedure FreeConverters;
  begin
    FreeAndNil( LocConverters);
  end;

  procedure SetSystemRate( aValue: Integer);
  begin
    System_Rate     := aValue;                        // samples / s
    System_Rate_2   := System_Rate * 2;               // samples / s
    Slow_Decimation := 8;
    Slow_Rate       := System_Rate / Slow_Decimation; // samples / s
    ComputeTables;
  end;


  procedure InitializeUnit;
  begin
    ReferenceA      := 440.0;
    NotesPerOctave  := 12.0;
    MiddleNote      := 69;
    SetSystemRate( 44100);
  end;


initialization

  CreateConverters;
  InitializeUnit;

finalization

  FreeConverters;

end.

