unit scala;

interface

uses

  System.Classes, System.SysUtils, System.Math, Rationals, Globals;


type

  EScala           = Exception;
  EScalaReject     = class( EScala);

  TScalaNoteType   = ( stRational, stCents);
  TNoteNr          = 0 .. 127;

  TScalaNoteValue = record
    NoteType : TScalaNoteType;
    case TScalaNoteType of
      stRational    : ( Rational: TRational);
      stCents       : ( Cents   : Double   );
   {end}
  end;

  TScalaTuningTable = array[ TNoteNr] of TScalaNoteValue;

  TScalaMappingFile = class( TStringList)
  strict private
    type TParseState = (
      msInitial,
      msHasMapSize,
      msHasFirstNote,
      msHasLastNote,
      msHasMiddleNote,
      msHasReferenceNote,
      msHasReferenceFrequency,
      msHasScaleDegree
    );
  private
    FFileName           : string;           // File name for file  that the mapping was read from
    FParseState         : TParseState;      // The current file parse state
    FMapSize            : Integer;          // Size of map. The pattern repeats every so many keys:
    FFirstNote          : Integer;          // First MIDI note number to retune
    FLastNote           : Integer;          // Last  MIDI note number to retune
    FMiddleNote         : Integer;          // Middle note where the first entry of the mapping is mapped to
    FReferenceNote      : Integer;          // Reference note for which frequency is given
    FReferenceFrequency : Double;           // Frequency to tune the above note to (Hz)
    FScaleDegree        : Integer;          // Scale degree to consider as formal octave (determines difference in pitch
                                            // between adjacent mapping patterns)
    FMapping            : array of Integer; // The numbers represent scale degrees mapped to keys. The first entry is
                                            // for the given middle note, the next for subsequent higher keys.
                                            // For an unmapped key, put in an "x". At the end, unmapped keys may be left
                                            // out.
  private
    function    GetMapSize: Integer;
    function    GetMapping( anIndex: Integer): Integer;
  private
    procedure   Reject( const aMsg: string);
    procedure   RejectFmt( const aFmt: string; anArgs: array of const);
    procedure   AddMapping( aLineNr: Integer; aValue: Integer);
    procedure   ParseLine( aLineNr: Integer; const aLine: string);
    procedure   Parse;
    procedure   SetDefaultMapping;
  public
    constructor Create( const aFileName: string);
  public
    property    FileName                    : string  read FFileName;
    property    MapSize                     : Integer read GetMapSize;
    property    FirstNote                   : Integer read FFirstNote;
    property    LastNote                    : Integer read FLastNote;
    property    MiddleNote                  : Integer read FMiddleNote;
    property    ReferenceNote               : Integer read FReferenceNote;
    property    ReferenceFrequency          : Double  read FReferenceFrequency;
    property    ScaleDegree                 : Integer read FScaleDegree;
    property    Mapping[ anIndex: Integer]  : Integer read GetMapping;
  end;


  TScalaScaleFile = class( TStringList)
  strict private
    type TParseState = (
      spInitial,
      spHasDescription,
      spHasNoteCount
    );
  private
    FFileName    : string;
    FparseState  : TParseState;
    FDescription : string;
    FNoteCount   : Integer;
    FNotes       : array of TScalaNoteValue;
    FTable       : TScalaTuningTable;
  private
    function    GetNoteCount: Integer;
    function    GetNote( anIndex: Integer): TScalaNoteValue;
    function    GetNoteType( anIndex: Integer): TScalaNoteType;
    function    GetRational( anIndex: Integer): Trational;
    function    GetCents   ( anIndex: Integer): Double;
  private
    procedure   Reject( const aMsg: string);
    procedure   RejectFmt( const aFmt: string; anArgs: array of const);
    procedure   AddRational( aNumerator: Int64; aDenominator: Int64 = 1);
    procedure   AddCents( aValue: Double);
    procedure   ParseCents   ( aLineNr: Integer; const aLine: string);
    procedure   ParseRational( aLineNr: Integer; const aLine: string);
    procedure   ParseLine    ( aLineNr: Integer; const aLine: string);
    procedure   FillTable;
    procedure   Parse;
  public
    constructor Create( const aFileName: string);
    function    TranslateNoteNr( aNoteNr: TNoteNr): Double;
  public
    property    FileName                    : string          read FFileName;
    property    NoteCount                   : Integer         read GetNoteCount;
    property    Note    [ anIndex: Integer] : TScalaNoteValue read GetNote;
    property    NoteType[ anIndex: Integer] : TScalaNoteType  read GetNotetype;
    property    Rational[ anIndex: Integer] : TRational       read GetRational;
    property    Cents   [ anIndex: Integer] : Double          read GetCents;
  end;



implementation


{ ========
  TScalaMappingFile = class( TStringList)
  strict private
    type TParseState = (
      msInitial,
      msHasMapSize,
      msHasFirstNote,
      msHasLastNote,
      msHasMiddleNote,
      msHasReferenceNote,
      msHasReferenceFrequency,
      msHasScaleDegree
    );
  private
    FFileName           : string;
    FParseState         : TParseState;
    FMapSize            : Integer;
    FFirstNote          : Integer;
    FLastNote           : Integer;
    FMiddleNote         : Integer;
    FReferenceNote      : Integer;
    FReferenceFrequency : Double;
    FScaleDegree        : Integer;
    FMapping            : array of Integer;
  public
    property    FileName                    : string  read FFileName;
    property    MapSize                     : Integer read GetMapSize;
    property    FirstNote                   : Integer read FFirstNote;
    property    LastNote                    : Integer read FLastNote;
    property    MiddleNote                  : Integer read FMiddleNote;
    property    ReferenceNote               : Integer read FReferenceNote;
    property    ReferenceFrequency          : Double  read FReferenceFrequency;
    property    ScaleDegree                 : Integer read FScaleDegree;
    property    Mapping[ anIndex: Integer]  : Integer read GetMapping;
  private
}

    function    TScalaMappingFile.GetMapSize: Integer;
    begin
      Result := Length( FMapping);
    end;

    function    TScalaMappingFile.GetMapping( anIndex: Integer): Integer;
    begin
      if ( anIndex >= 0) and ( anIndex < MapSize)
      then Result := FMapping[ anIndex]
      else Result := -1;
    end;

//  private

    procedure   TScalaMappingFile.Reject( const aMsg: string);
    begin
      raise EScalaReject.Create( aMsg);
    end;

    procedure   TScalaMappingFile.RejectFmt( const aFmt: string; anArgs: array of const);
    begin
      Reject( Format( aFmt, anArgs));
    end;

    procedure   TScalaMappingFile.AddMapping( aLineNr: Integer; aValue: Integer);
    begin
      if FMapSize > MapSize
      then begin
        SetLength( FMapping, MapSize + 1);
        FMapping[ MapSize - 1] := aValue;
      end
      else RejectFmt( 'Too many Mappings for given Map Size in line %d', [ aLineNr]);
    end;

    procedure   TScalaMappingFile.ParseLine( aLineNr: Integer; const aLine: string);
    var
      N : Integer;
      F : Double;
    begin
      if ( Pos( '!', aLine) <> 1) and ( aLine <> '')
      then begin
        case FparseState of

          msInitial :
            begin
              N := StrToIntDef( aLine, -1);
              if N >= 0
              then begin
                FMapSize := N;
                FparseState := msHasMapSize;
              end
              else RejectFmt( 'Invalid map size in line %d', [ aLineNr]);
            end;

          msHasMapSize :
            begin
              N := StrToIntDef( aLine, -1);
              if N >= 0
              then begin
                FFirstNote := N;
                FparseState := msHasFirstNote;
              end
              else RejectFmt( 'Invalid first note in line %d', [ aLineNr]);
            end;

          msHasFirstNote :
            begin
              N := StrToIntDef( aLine, -1);
              if N >= 0
              then begin
                FLastNote := N;
                FparseState := msHasLastNote;
              end
              else RejectFmt( 'Invalid last note in line %d', [ aLineNr]);
            end;

          msHasLastNote :
            begin
              N := StrToIntDef( aLine, -1);
              if N >= 0
              then begin
                FMiddleNote := N;
                FparseState := msHasMiddleNote;
              end
              else RejectFmt( 'Invalid middle note in line %d', [ aLineNr]);
            end;

          msHasMiddleNote :
            begin
              N := StrToIntDef( aLine, -1);
              if N >= 0
              then begin
                FReferenceNote := N;
                FparseState := msHasReferenceNote;
              end
              else RejectFmt( 'Invalid reference note in line %d', [ aLineNr]);
            end;

          msHasReferenceNote :
            begin
              F := StrToFloatDef( aLine, -1);
              if F >= 0
              then begin
                FReferenceFrequency := F;
                FparseState := msHasReferenceFrequency;
              end
              else RejectFmt( 'Invalid reference frequency in line %d', [ aLineNr]);
            end;

          msHasReferenceFrequency :
            begin
              N := StrToIntDef( aLine, -1);
              if N >= 0
              then begin
                FScaleDegree := N;
                FparseState := msHasScaleDegree;
              end
              else RejectFmt( 'Invalid scale degree in line %d', [ aLineNr]);
            end;

          msHasScaleDegree :
            begin
              N := StrToIntDef( aLine, -1);
              if N >= 0
              then AddMapping( aLineNr, N)
              else RejectFmt( 'Invalid mapping in line %d', [ aLineNr]);
            end;

        end;
      end;
    end;

    procedure   TScalaMappingFile.Parse;
    var
      i : Integer;
    begin
      SetLength( FMapping, 0);
      FparseState := msInitial;
      for i := 0 to Count - 1
      do ParseLine( i + 1, Trim( Strings[ i]));
      if FMapSize <> MapSize
      then Reject( 'Given Map Size does not match the actual number of Mappings found');
    end;

    procedure   TScalaMappingFile.SetDefaultMapping;
    var
      i : Integer;
    begin
      FMapSize            := 128;
      FFirstNote          :=   0;
      FLastNote           := 127;
      FMiddleNote         :=  60;
      FReferenceNote      :=  69;
      FReferenceFrequency := 440;
      FScaleDegree        :=  12;
      SetLength( FMapping, 128);
      for i := 0 to 127
      do FMapping[ i] := i;
    end;

//  public

    constructor TScalaMappingFile.Create( const aFileName: string);
    begin
      inherited Create;
      FFileName := aFileName;
      if FileExists( FFileName)
      then begin
        LoadFromFile( FFileName);
        Parse;
      end
      else begin
        if FFileName = ''
        then SetDefaultMapping
        else RejectFmt( 'mapping file ''%s'' not found', [ FFileName])
      end;
    end;


{ ========
  TScalaScaleFile = class( TStringList)
  private
    type TScalaParseState = ( spInitial, spHasDescription, spHasNoteCount);
  private
    FFileName    : string;
    FparseState  : TScalaParseState;
    FDescription : string;
    FNoteCount   : Integer;
    FNotes       : array of TScalaNoteValue;
    FTable       : TScalaTuningTable;
  public
    property    NoteCount                   : Integer         read GetNoteCount;
    property    Note    [ anIndex: Integer] : TScalaNoteValue read GetNote;
    property    NoteType[ anIndex: Integer] : TScalaNoteType  read GetNotetype;
    property    Rational[ anIndex: Integer] : TRational       read GetRational;
    property    Cents   [ anIndex: Integer] : Double          read GetCents;
  private
}

    function    TScalaScaleFile.GetNoteCount: Integer;
    begin
      Result := Length( FNotes);
    end;

    function    TScalaScaleFile.GetNote( anIndex: Integer): TScalaNoteValue;
    begin
      Result := FNotes[ anIndex];
    end;

    function    TScalaScaleFile.GetNoteType( anIndex: Integer): TScalaNoteType;
    begin
      Result := FNotes[ anIndex].NoteType;
    end;

    function    TScalaScaleFile.GetRational( anIndex: Integer): Trational;
    begin
      if NoteType[ anIndex] = stRational
      then Result := FNotes[ anIndex].Rational
      else Result := TRational.Create( FNotes[ anIndex].Cents); // todo : convert cemts to rational
    end;

    function    TScalaScaleFile.GetCents( anIndex: Integer): Double;
    begin
      if NoteType[ anIndex] = stCents
      then Result := FNotes[ anIndex].Cents
      else Result := FNotes[ anIndex].Rational.AsDouble; // todo : convert rational to cents
    end;

//  private

    procedure   TScalaScaleFile.Reject( const aMsg: string);
    begin
      raise EScalaReject.Create( aMsg);
    end;

    procedure   TScalaScaleFile.RejectFmt( const aFmt: string; anArgs: array of const);
    begin
      Reject( Format( aFmt, anArgs));
    end;

    procedure   TScalaScaleFile.AddRational( aNumerator: Int64; aDenominator: Int64 = 1);
    begin
      SetLength( FNotes, Length( FNotes) + 1);
      FNotes[ NoteCount - 1].NoteType := stRational;
      FNotes[ NoteCount - 1].Rational := TRational.Create( aNumerator, aDenominator)
    end;

    procedure   TScalaScaleFile.AddCents( aValue: Double);
    begin
      SetLength( FNotes, Length( FNotes) + 1);
      FNotes[ NoteCount - 1].NoteType := stCents;
      FNotes[ NoteCount - 1].Cents    := aValue;
   end;

    procedure   TScalaScaleFile.ParseCents( aLineNr: Integer; const aLine: string);
    var
      N : Double;
    begin
      N := StrToFloatDef( aLine, NaN);
      if IsNan( N) or ( N < 0)
      then RejectFmt( 'Invalid cents spec in line %d', [ aLineNr])
      else AddCents( N);
    end;

    procedure   TScalaScaleFile.ParseRational( aLineNr: Integer; const aLine: string);
    var
      S      : string;
      p      : Integer;
      aParts : TStringList;
      N1     : Integer;
      N2     : Integer;
      aCode  : Integer;
    begin
      p := Pos( ' ', aLine);
      if p <= 0
      then p := Length( aLine);
      S := Copy( aLine, 1, p);
      aParts := Explode( S, '/');
      try
        if aParts.Count = 1
        then begin
          Val( aParts[ 0], N1, aCode);
          if aCode = 0
          then AddRational( N1)
          else RejectFmt( 'invalid note in line %d', [ aLineNr]);
        end
        else if aParts.Count = 2
        then begin
          Val( aParts[ 0], N1, aCode);
          if aCode = 0
          then begin
            Val( aParts[ 1], N2, aCode);
            if aCode = 0
            then AddRational( N1, N2)
            else RejectFmt( 'invalid note in line %d', [ aLineNr]);
          end
          else RejectFmt( 'invalid note in line %d', [ aLineNr]);
        end
        else RejectFmt( 'invalid note in line %d', [ aLineNr]);
      finally
        aParts.Free;
      end;
    end;

    procedure   TScalaScaleFile.ParseLine( aLineNr: Integer; const aLine: string);
    var
      N : Integer;
    begin
      if ( Pos( '!', aLine) <> 1) and ( aLine <> '')
      then begin
        case FparseState of

          spInitial :
            begin
              FDescription := aLine;
              FparseState  := spHasDescription;
            end;

          spHasDescription :
            begin
              N := StrToIntDef( aLine, -1);
              if N >= 0
              then begin
                FNoteCount := N;
                AddRational( 1);
                FparseState := spHasNoteCount;
              end
              else RejectFmt( 'Invalid note count in line %d', [ aLineNr]);
            end;

          spHasNoteCount :
            begin
              if Pos( '.', aLine) > 0
              then ParseCents   ( aLineNr, aLine)
              else ParseRational( aLineNr, aLine);
            end;

        end;
      end;
    end;

    procedure   TScalaScaleFile.FillTable;
    begin
      // todo : fill the translation table
    end;

    procedure   TScalaScaleFile.Parse;
    var
      i : Integer;
    begin
      SetLength( FNotes, 0);
      FparseState := spInitial;
      for i := 0 to Count - 1
      do ParseLine( i + 1, Trim( Strings[ i]));
      if FNoteCount <> NoteCount - 1
      then Reject( 'Given note count does not match the actual number of note specifications found');
      FillTable;
    end;

//  public

    constructor TScalaScaleFile.Create( const aFileName: string);
    begin
      inherited Create;
      FFileName := aFileName;
      if FileExists( FFileName)
      then begin
        LoadFromFile( FFileName);
        Parse;
      end
      else RejectFmt( 'File ''%s'' not found', [ FFileName]);
    end;

    function    TScalaScaleFile.TranslateNoteNr( aNoteNr: TNoteNr): Double;
    begin
      if FTable[ aNoteNr].NoteType = stCents
      then Result := FTable[ aNoteNr].Cents
      else Result := FNotes[ aNoteNr].Rational.AsDouble; // todo : convert rational to cents
    end;


end.

